function [m,output]=ec_fac_ep(theta,J,par,debug_draw)
% ec_fac_ep.m 
% expectation consistent factorized inference
% for binary variables
% optimized using ep-style updates 
% uses ec_fac_grad when ep doesn't convergence

% Ole Winther, IMM, February 2005
  
% initialize variables
N = length(theta); 

% initialize constants
try ftol=par.ftol; catch ftol = 1e-7; end  
try minchi=par.minchi; catch minchi = 1e-5; end
try m=par.m; catch m=zeros(N,1); end % initial value of m
try max_ite=par.max_ite; catch max_ite=100; end
try eta=par.eta; catch eta=0.5; end % set learning rate for lam_r update
try draw = par.draw; catch draw = 0; end
try G = par.G; catch G = 0; end
try history = par.history; catch history = 0; end

if draw==1,fprintf('** Start ec_fac_ep ************************************ N=%d\n',N);tic;end

dm = zeros(N,1); m_r = zeros(N,1); v_r = zeros(N,1);

try Lam_r = par.Lam_r; 
catch
  % make sure that chi = inv( diag(Lam_r) - J ) is positive definite
  eigJ=eig(J); mineigJ = min(eigJ); maxeigJ = max(eigJ);
  Lam_r = 0*maxeigJ + 2 * ( maxeigJ - mineigJ ) * ones(N,1) ;
end
try chi = par.chi; catch chi =  inv( diag(Lam_r) - J ) ; end

% set mean value of r-distribution to be the same as m
gam_r = zeros(N,1) ; % ( diag(Lam_r) - J ) * m - theta ;
m_r = m ;
v_r = diag(chi) ;

dmdm=inf; dmdm_old=Inf; ite=0; overcount = 0;
while dmdm>ftol & ite<max_ite 
  ite=ite+1;
  indx=randperm(N);
  for sindx=1:N
    cindx=indx(sindx);

    %% find mean and variance of r
    m_r(cindx) = chi(cindx,:) * ( gam_r + theta ) ;
    v_r(cindx) = chi(cindx,cindx);
    
    % find Lagrange parameters of s 
    Lam_s1 = 1 / max( minchi , v_r(cindx) ) ;
    gam_s1 = Lam_s1 * m_r(cindx); 
       
    % update lam_q
    gam_q = gam_s1 - gam_r(cindx) ; 
    Lam_q = Lam_s1 - Lam_r(cindx) ;

    % update moments of q distribution
    m(cindx) = tanh( gam_q );
    v_q = 1 - m(cindx)^2 ;

    dm(cindx) = m(cindx) - m_r(cindx) ;

    % find Lagrange parameters of s 
    Lam_s2 = 1 / max( minchi , v_q ) ;
    gam_s2 = Lam_s2 * m(cindx) ;

    % update lam_r
    dgam_r = eta * ( gam_s2 - gam_s1 ) ; gam_r(cindx) = dgam_r + gam_r(cindx) ;
    dLam_r = eta * ( Lam_s2 - Lam_s1 ) ; Lam_r(cindx) = dLam_r + Lam_r(cindx) ;

    % update chi using Sherman-Woodbury
    chiv = chi(:,cindx); 
    chi = chi  - dLam_r / ( 1 + dLam_r * v_r(cindx) ) * chiv * chiv';  % v_r = chi(cindx,cindx)
    
  end % over variables - sindx 
  
  dmdm = dm' * dm; 
  
  if dmdm > dmdm_old % lower learning rate and start over
    overcount = overcount + 1;
    if overcount == 50; % switch to another method
      if draw==1,fprintf('** ec_fac_ep doesn''t seem to convergence **** %4.3f sec ****\n',toc);end
      %[m,output]=ec_fac_seqc(theta,J,par);
      output.converged=0;
      output.dmdm=dmdm;
      output.ite=ite;
      return
    end
  else
    overcount = 0;  
    dmdm_old=dmdm;
  end
  
  if history
  
    output.m(:,ite) = m;
    output.m_r(:,ite) = m_r;
    output.v_r(:,ite) = v_r;

  end

end

if ite == max_ite
  if draw==1,fprintf('** ec_fac_ep didn''t converge in %d iterations * %4.3f sec **\n',ite,toc);end
  output.ite = ite ;
  %[m,output]=ec_fac_cv(theta,J,par);
  output.dmdm=dmdm;
  output.converged=0;
  return
end

if G % calculate free energy
  gam_q = m ./ max( minchi , 1 - m.^2 ) - gam_r ; % gam_s - gam_r 
  G_q = - sum( -log(2) + abs( gam_q ) + log( 1 + exp( -2 * abs( gam_q ) ) ) ) + m' * gam_q ;
  G_r = 0.5 * ( m' * ( gam_r + theta ) - logdet(chi) - sum(Lam_r) ) - m' * theta ;
  G_s = - 0.5 * ( sum( log( max( minchi, 1 - m .* m ) ) ) + N ) ;
  output.G = G_q + G_r - G_s ;
end

if nargout > 1
  output.chi = chi;  
  output.converged = 1;
  output.ite = ite ;
  output.dmdm = dmdm;
  output.gam_q = m ./ max( minchi , 1 - m.^2 ) - gam_r ; % gam_s - gam_r 
  output.gam_r = gam_r ;
  output.Lam_r = Lam_r ;
end

if draw==1,fprintf('** End of ec_fac_ep ****** time %4.3f sec********** ite =%3d\n\n',toc,ite);end

function y = logdet(A)
% log(det(A)) where A is positive-definite.
% This is faster and more stable than using log(det(A)).

U = chol(A);
y = 2*sum(log(diag(U)));
