function [m,output]=ec_fac_jtree(theta,J,par,debug_draw)
% ec_fac_seqc.m 
% expectation consistent factorized inference
% for binary variables
% optimized using sequential variational bounding
% "guarenteed convergent" 

% Ole Winther, IMM, February 2005
  
% initialize variables
N = length(theta); 

% initialize constants
try ftol=par.ftol; catch ftol = 1e-7; end  
try minchi=par.minchi; catch minchi = 1e-5; end
try m=par.m; catch m=zeros(N,1); end % initial value of m
try max_ite=par.max_ite; catch max_ite=100; end
try eta=par.eta; catch eta=1; end % set learning rate for lam_r update
try draw = par.draw; catch draw = 0; end
try G = par.G; catch G = 0; end
try history = par.history; catch history = 0; end
try ro = par.ro; catch ro = 1; end % toggle random order 

if draw==1,fprintf('** Start ec_fac_jtree ************************************ N=%d\n',N);tic;end

output.converged = 1 ;

dm = zeros(N,1); m_r = zeros(N,1); v_r = zeros(N,1); 

try Lam_r = par.Lam_r; 
catch
  % make sure that chi = inv( diag(Lam_r) - J ) is positive definite
  eigJ=eig(J); mineigJ = min(eigJ); maxeigJ = max(eigJ);
  Lam_r = 0*maxeigJ + 2 * ( maxeigJ - mineigJ ) * ones(N,1) ;
end
try chi = par.chi; catch chi =  inv( diag(Lam_r) - J ) ; end

% set mean value of r-distribution to be the same as m
gam_r = zeros(N,1); %( diag(Lam_r) - J ) * m - theta ; 
%m_r = m ; v_r = diag( chi ) ;

% init lam_q
%gam_q = - gam_r; Lam_q = - Lam_r;

% init lam_s
%gam_s2 = zeros(N,1); Lam_s2 = zeros(N,1) ;

m_best = zeros(N,1) ; dmdm_best = inf ;

dmdm=inf; dmdm_old=Inf; ite=0; overcount = 0;

while dmdm>ftol & ite<max_ite 
    
  ite=ite+1;
  
  % distribute evidence
  
  % find mean and variance of r
  m_r = chi * ( gam_r + theta ) ;
  v_r = diag( chi ) ;
          
  %dm = m - m_r ;
  %dmdm1 = dm' * dm 
   
  % find Lagrange parameters of s 
  Lam_s1 = 1 ./ max( minchi , v_r ) ;
  gam_s1 = Lam_s1 .* m_r; 
       
  % update lam_q
  gam_q = gam_s1 - gam_r ; % eta * ( gam_s1 - gam_s2 ) + gam_q ; 
  Lam_q = Lam_s1 - Lam_r ; %eta * ( Lam_s1 - Lam_s2 ) + Lam_q ;

  % collect evidence
  
  % update moments of q distribution
  m = tanh( gam_q );
  v_q = 1 - m.^2 ; 

  % break if NAN
  if any(isnan(m))
    m = m_best ;
    if draw==1,
      fprintf('** Does not converge, return solution with dmdm = %g **\n',dmdm_best);
    end
    output.converged = 0 ; 
    break ;      
  end
  
  dm = m - m_r ;
   
  dmdm = dm' * dm ;
  
  % find Lagrange parameters of s and update lam_s_star
  Lam_s2 = 1 ./ max( minchi , v_q ) ;
  gam_s2 = Lam_s2 .* m ;

  % update lam_r
  dgam_r = eta * ( gam_s2 - gam_s1 ) ; gam_r = dgam_r + gam_r ;
  dLam_r = eta * ( Lam_s2 - Lam_s1 ) ; Lam_r = dLam_r + Lam_r ;

  % update chi
  chi = inv ( diag(Lam_r) - J ) ; 
     
  %mineigchi=min(eig(chi))
  
  if dmdm < dmdm_best 
    m_best = m ;
    dmdm_best = dmdm ;
  end
  
  if history
    output.m(:,ite) = m;
    output.m_r(:,ite) = m_r;
    output.v_r(:,ite) = v_r;
  end

end % while

if ite == max_ite output.convergence=0; else output.convergence=1; end

if G % calculate free energy
  gam_q = m ./ max( minchi , 1 - m.^2 ) - gam_r ; % gam_s - gam_r 
  G_q = - sum( -log(2) + abs( gam_q ) + log( 1 + exp( -2 * abs( gam_q ) ) ) ) + m' * gam_q ;
  G_r = 0.5 * ( m' * ( gam_r + theta ) - logdet(chi) - sum(Lam_r) ) - m' * theta ;
  G_s = - 0.5 * ( sum( log( max( minchi, 1 - m .* m ) ) ) + N ) ;
  output.G = G_q + G_r - G_s ;
end

if nargout > 1
  output.chi = chi;  
  output.dmdm = dmdm;
  output.gam_q = m ./ max( minchi , 1 - m.^2 ) - gam_r ; % gam_s - gam_r 
  output.gam_r = gam_r ;
  output.ite = ite ;
  output.Lam_r = Lam_r ;
end
  
if draw==1,fprintf('** End of ec_fac_jtree ****** time %4.3f sec********** ite =%3d\n\n',toc,ite);end

function y = logdet(A)
% log(det(A)) where A is positive-definite.
% This is faster and more stable than using log(det(A)).

U = chol(A);
y = 2*sum(log(diag(U)));
