function [m,output]=ec_tree_jtree(theta,J,par,debug_draw)
% ec_tree_jtree
% expectation consistent tree structured inference
% for binary variables
% optimized parallel junction tree style

% Ole Winther, IMM, March 2005
%
%
% I made some changes: W. Wiegerinck SNN, APRIL 2005
% Basically I changed the update per edge to an updat of all edges per 
% iteration, to gain a speed up. Means seem the same, correlations is not so clear.
  
%warning off all % might not converge because of singular matrices

% initialize variables
N = length(theta); 

% initialize constants
try ftol=par.ftol; catch ftol = 1e-7; end  
try minchi=par.minchi; catch minchi = 1e-5; end
try m=par.m; catch m=zeros(N,1); end % initial value of m
try max_ite=par.max_ite; catch max_ite=100; end
try eta=par.eta; catch eta=0.5; end % set learning rate for lam_r update
try draw = par.draw; catch draw = 0; end
try G = par.G; catch G = 0; end
try history = par.history; catch history = 0; end
try treeop = par.treeop; catch treeop='J'; end
switch lower(treeop)
  case 'j'
    [nodes,ncon,CM] = find_tree(abs(J)); % use maximal spanning tree abs(J)
  case 'j_pairs'
    [nodes,ncon] = find_pairs(abs(J)); % use large abs(J) to set connectivity
  case 'random'
    [nodes,ncon,CM] = find_tree(rand(N,N)); % random graph
  case 'correlation'
    [nodes,ncon,CM] = find_tree(par.Ccorr); % use predefined correlation matrix
  otherwise
    disp('unknown treeop')  
end   

links = prop_flow(CM, 1);
output.converged = 1 ;

if draw==1,fprintf('** Start ec_tree_jtree ************************************ N=%d\n',N);tic;end

try Lam_r = par.Lam_r; 
catch
  % make sure that chi = inv( diag(Lam_r) - J ) is positive definite
  eigJ=eig(J); mineigJ = min(eigJ); maxeigJ = max(eigJ);
  Lam_r = 0*maxeigJ + 2 * ( maxeigJ - mineigJ + eps) * ones(N,1) ;
end
Lam_r = sparse(1:N,1:N,Lam_r);
try chi_r = par.chi; catch chi_r =  inv( full(Lam_r - J) ) ; end

% set mean vale of r-distribution 
gam_r = zeros(N,1) ;  % ( Lam_r - J ) * m - theta ;

Nnodes = length(nodes);

dm = zeros(N,1);
  
mm = zeros(Nnodes,N); 

m_best = zeros(N,1) ; dmdm_best = inf ;

dmdm=inf; ite=0;
while dmdm>ftol & ite<max_ite 
  ite=ite+1;
  %iperm = randperm(Nnodes);
 
  % set mean of r
  m_r = chi_r * ( gam_r + theta );
  
  % solve for separator s
  [gam_s1,Lam_s1] = treegauss(links,m_r,chi_r);

  % set lam_q
  gam_q = gam_s1 - gam_r;
  Lam_q = Lam_s1 - Lam_r;
  
  % set mean and covariance of q
  [m,chi_q] = treebp(links,gam_q,-Lam_q);
  
  % break if NAN
  if any(isnan(m))
    m = m_best ;
    if draw==1,
      fprintf('** Does not converge, return solution with dmdm = %g **\n',dmdm_best);
    end
    output.converged = 0 ; 
    break ;      
  end

  dm = m - m_r ;
  
  % solve for separator
  [gam_s2,Lam_s2] = treegauss(links,m,chi_q);

  % set lam_r
  dgam_r = eta * ( gam_s2 - gam_s1 ) ; 
  gam_r = dgam_r + gam_r ;
  dLam_r = eta * ( Lam_s2 - Lam_s1 ) ; 
  Lam_r = dLam_r + Lam_r ;
  
  % update covariance of r
  chi_r =  inv( Lam_r - J );
    
  dmdm = dm' * dm ; 
  
  if dmdm < dmdm_best 
    m_best = m ;
    dmdm_best = dmdm ;
  end
   
end

if G % calculate free energy
  output.G = Ftree(m,chi_r,Lam_r-J,Lam_q,nodes,ncon) ; % inv(chi_r)=Lam_r-J;
end

if nargout > 1
  output.chi = chi_r;  
  output.dmdm = dmdm;
  output.gam_q = gam_q ;
  output.gam_r = gam_r ;
  output.Lam_r = Lam_r ;
  output.ite = ite ;
  output.mm = mm ; 
end

if draw==1,fprintf('** End of ec_tree_jtree ****** time %4.3f sec********** ite =%3d\n\n',toc,ite);end


function [links,ncon,CM] = find_tree(C,ncon)
% constructs a tree from the 
% non-negative N*N correlation matrix C

N = size(C,1);

try ncon; catch ncon=N-1; end
pairs = 0;

[Y,I] = sort(C(:));
I = flipud(I);

indx = 1;

graph = (1:N)';

CM=sparse(N,N);

while pairs < ncon % we are looking for N-1 pairs

    i = fix( ( I(indx) - 1 ) / N ) + 1;
    j = rem( I(indx)-1 , N ) + 1; 

    if (graph(i) ~= graph(j)) % make link if nodes does not belong to the same sub-graph

        pairs = pairs + 1 ;

        links{pairs}.indx = [ i j ] ;
        
        graph(find(graph == graph(j))) = graph(i); % relabel nodes    

        CM(i,j) = 1;

    end
    
    indx = indx + 1;
    
end

% set adjacency matraix to be symmetric
CM = CM + CM';

% for each link find adjacent links - probably not optimal code
for i=1:ncon
   
  i1=links{i}.indx(1);
  i2=links{i}.indx(2);
  k1=0; k2=0;  links{i}.links1 = [];  links{i}.links2 = [] ;
  for j=[1:i-1 i+1:ncon]
    if any(links{j}.indx == i1)
      k1 = k1 + 1;
      links{i}.links1(k1) = j ;
    end
    if any(links{j}.indx == i2)
      k2 = k2 + 1;
      links{i}.links2(k2) = j ;
    end
  end 
end

% find order of transversal for all subtrees rooted in each link
for i=1:ncon
  pindx = 0 ; 
  CMred=CM; CMred( links{i}.indx(1) , links{i}.indx(2) ) = 0 ;  
  CMred( links{i}.indx(2) , links{i}.indx(1) ) = 0 ;
  for parent = links{i}.indx     
    pindx = pindx + 1 ; 
    k = 1 ; level = 0 ;  
    A = CMred(parent,:) ; indx = find( A ) ; B = CMred ; B(:,parent) = 0 ;  
    while ~isempty(indx)
   
      level = level + 1 ;
    
      [ii,jj,s] = find( sparse(parent,parent,1,N,N) * B  ) ; 
      tmp(pindx).nlinks(level) = length(s) ; 
      tmp(pindx).linkorder(k:k+tmp(pindx).nlinks(level)-1,:) = [ ii jj ] ;
      k = k + tmp(pindx).nlinks(level) ;
      parent = indx ;
  
      B(:,indx) = 0 ; % to avoid going backwards
  
      A = A * B ;
      indx = find(A) ;
  
    end
    tmp(pindx).levels = level ;
    if level tmp(pindx).linkorder = flipud(tmp(pindx).linkorder) ; end
  end % parents
  links{i}.levels1 = tmp(1).levels ;
  links{i}.levels2 = tmp(2).levels ;
  if links{i}.levels1 
    links{i}.lo1 = tmp(1).linkorder; 
    links{i}.nlinks1 = tmp(1).nlinks; 
  else links{i}.lo1 = [] ; links{i}.nlinks1 = [] ; end
  if links{i}.levels2
    links{i}.lo2 = tmp(2).linkorder; 
    links{i}.nlinks2 = tmp(2).nlinks; 
  else links{i}.lo2 = [] ; links{i}.nlinks2 = [] ; end
  clear tmp
end % ncon



%graph
ncon = (sum(CM,1))';    


if 0
k = 1 ;
parent = 1 ;
level = 0 ; 
A = CM(parent,:) ; indx = find( A ) ; B = CM ; B(:,parent) = 0 ; 
while ~isempty(indx)

  [i,j,s] = find( sparse(parent,parent,1,N,N) * B  )   ;
  nlinks = length(s) ; 
  linkorder(k:k+nlinks-1,:) = [ i j ] ;
  k = k + nlinks ;
  parent = indx ;
  
  B(:,indx) = 0 ; % to avoid going backwars
  level = level + 1;
  cl(level).indx = indx ; 
  
  A = A * B ;
  indx = find(A) ;
  
end
end  
  
plotgraph=0;
if plotgraph 
    gplot(CM,[-cos(2*pi*(0:N-1)'/N) sin(2*pi*(0:N-1)'/N)],'o-'), hold on
    plot(-cos(2*pi*(0:N-1)'/N),sin(2*pi*(0:N-1)'/N),'*'), hold off, drawnow
    bigfig(2)
end
plotgraphgrid=0;
if plotgraphgrid 
    sqrtN=sqrt(N); % assume integer
    gplot(CM,0.125+[mod((0:N-1)',sqrtN)/sqrtN floor((0:N-1)'/sqrtN)/sqrtN],'o-'), hold on
    plot(0.125+mod((0:N-1)',sqrtN)/sqrtN,0.125+floor((0:N-1)'/sqrtN)/sqrtN,'*'), 
    hold off, drawnow
    bigfig(2)
end

function [f] = prop_flow(adj, root)
%
% makes a propagation order for a adjacency matrix adj, starting in root
% adj is a sparse nxn connection matrix (assumed to be a tree)
%
%

try root=root; catch root =1; end;

N = size(adj,1);
depth = zeros(N,1);
depth(root) = 1;
queue = root;

f = [];

while ~isempty(queue),
  r = queue(1);
  queue(1) = [];
  idx = find(adj(r,:));
  idx2 = find(~depth(idx));
  idx = idx(idx2);
  queue = [queue idx];
  depth(idx) = depth(r)+1;
  f = [f ; [repmat(r,size(idx(:))) idx(:)]];
end;

function [h,K] = treegauss(link,m,sigma)
%
% Computes paramters of a Gauss - tree given sigma and mu
%
nlink = size(link,1);
n = nlink+1;
h  = zeros(n,1);
K = sparse(n,n);
nn = zeros(n,1);

for ij = 1:nlink,
  nn(link(ij,1)) = nn(link(ij,1))+1;
  nn(link(ij,2)) = nn(link(ij,2))+1;
    K(link(ij,:),link(ij,:)) = K(link(ij,:),link(ij,:)) + ...
	inv(sigma(link(ij,:),link(ij,:)));
  end;
  nnz = find(nn>1);
for i = nnz(:)'
  K(i,i) = K(i,i) - (nn(i)-1)/sigma(i,i);
end;

h = K*m;

function [m,sig] = treebp(link,theta,J)
%
% tree-bp uses link from link = prop_flow(J,root)
%
% outcome means <x> and <x^2>-<x>^2 on the edges;
%
%

nlink = size(link,1);
n = nlink+1;
messtofrom = sparse(n,n);
incomingexceptto = sparse(n,n);
sig = speye(n);
tanhJ = tanh(J);

for jj = nlink:-1:1,
  i = link(jj,1);
  j = link(jj,2);
  messtofrom(i,j) = incomingexceptto(i,j) + ...
      theta(j);
  messtofrom(i,j) = atanh(tanhJ(i,j) * ...
      tanh(messtofrom(i,j)));
  incomingexceptto(:,i) = incomingexceptto(:,i) + ones(n,1)*messtofrom(i,j);
end; 
incomingexceptto = incomingexceptto - messtofrom';

for jj = 1:nlink,
  i = link(jj,1);
  j = link(jj,2);
  messtofrom(j,i) = incomingexceptto(j,i) + ...
      theta(i);
  messtofrom(j,i) = atanh(tanhJ(j,i) * ...
      tanh(messtofrom(j,i)));
  receiver  = j;
  nextsteps = link(find(link(:,1) == receiver),2); 

  for nextstep = nextsteps(:)',
    incomingexceptto(nextstep,receiver) = incomingexceptto(nextstep,receiver) +...
	messtofrom(receiver,i);
  end;
end;

% means <xi>
m=tanh(theta +sum(messtofrom,2));

% sig = <xixj> - <xi><xj>
for jj = 1:nlink,
  i = link(jj,1);
  j = link(jj,2);
  mi=incomingexceptto(j,i)+theta(i);
  mj=incomingexceptto(i,j)+theta(j);
  w = J(i,j);
  rpp=exp(mi+mj+w);
  rpm =exp(mi-mj-w);
  rmp =exp(-mi+mj-w);
  rmm =exp(-mi-mj+w);
  sig(i,j)  = (rpp-rpm-rmp + rmm)/(rpp+rpm+rmp + rmm) - m(i)*m(j);
  sig(j,i) = sig(i,j);
end;

sig = sig - diag(m.^2);


function [F] = Ftree(m,chi_r,invchi_r,Lam_q,nodes,ncon)
% this function is only valid at convergence
% to look at each term independently one must define logZ_q and logZ_s in
% terms of their Lagrange parameters

N = length(m); 

logZ_r = 0.5 * ( m' * invchi_r * m + logdet(chi_r) ) ;

% single node contribution
logZ_q = ( 1 - ncon )' * log( 2*cosh( atanh( m ) ) ) - 0.5 * sum( diag( Lam_q ) ) ; 
logZ_s = 0.5 * ( 1 - ncon )' * ( log( diag( chi_r ) ) + m.^2 ./ diag( chi_r ) ) ;

Npair = length(nodes);
    
for i=1:Npair

  ip = nodes{i}.indx ;
 
  % update logZ_s
  logZ_s = logZ_s + 0.5 * logdet( chi_r(ip,ip) ) + 0.5 * m(ip)' * inv( chi_r(ip,ip) ) * m(ip) ; 
  
  % update logZ_q
  m1 = m(ip(1)) ; m2 = m(ip(2)) ;
  M = chi_r(ip(1),ip(2)) + m1 * m2;
  gam1 = 0.5 * atanh( ( m1 + m2 ) / ( 1 + M ) ) + 0.5 * atanh( ( m1 - m2 ) / ( 1 - M ) ) ;
  gam2 = 0.5 * atanh( ( m1 + m2 ) / ( 1 + M ) ) + 0.5 * atanh( ( m2 - m1) / ( 1 - M ) ) ;
  Lam_qij =  - 0.5 * atanh( ( m1 + M ) / ( 1 + m2 ) ) - 0.5 * atanh( ( M - m1 ) / ( 1 - m2 ) ) ;
  
  logZ_q = logZ_q + log( 2*cosh(gam1+gam2)*exp(-Lam_qij)+ 2*cosh(gam1-gam2)*exp(Lam_qij) ) ;
  
end

F = -logZ_q - logZ_r + logZ_s ;  

function y = logdet(A)
% log(det(A)) where A is positive-definite.
% This is faster and more stable than using log(det(A)).

U = chol(A);
y = 2*sum(log(diag(U)));
